import weakref
import pygame
from spw import paths


class Canvas(object):
    CACHE_SIZE = 4

    def __init__(self, surface):
        self.surface = surface
        self._x, self._y = 0.0, 0.0
        self._s = 1.0
        self._color = (255,255,255,255)
        self._line_width = 1
        self._state = []
        self._name = ()
        self._rects = {}
        self._dirty = False
        self._images = weakref.WeakKeyDictionary()
        self._blend_flags = 0

    def query(self, p, d=3):
        q = pygame.Rect(p[0], p[1], d, d)
        return q.collidedictall(self._rects)

    def _add_rect(self, r):
        key = tuple(r)
        names = self._rects.get(key, None)
        if names is None: names = self._rects[key] = set()
        if self._name: names.add(self._name)
    
    def tx(self, xy):
        x, y = xy
        return int(round(self._x + (x*self._s))), int(round(self._y + (y * self._s)))

    def un_tx(self, xy):
        x, y = xy
        return (x - self._x) / self._s, (y - self._y) / self._s

    def tx_points(self, points):
        X,Y,S = self._x, self._y, self._s
        return [(int(round((X + (x * S)))), int(round(Y + (y * S)))) for x,y in points]

    def clear(self):
        self.identity()
        self._name = ()
        self._rects = {}
        self._state = []
        self._blend_flags = 0

    def identity(self):
        self._x, self._y, self._s = 0.0, 0.0, 1.0
        
    def translate(self, xy):
        self._x, self._y = self.tx(xy)

    def scale(self, s):
        self._s *= s

    def color(self, rgba):
        self._color = rgba

    def blend(self, flags):
        self._blend_flags = flags 

    def line_width(self, w):
        self._line_width = int(w)

    def push_name(self, name):
        self._name += (name,)

    def pop_name(self):
        self._name = self._name[0:-1]

    def push(self):
        self._state.append((self._x, self._y, self._s, self._color, self._line_width))
        assert len(self._state) < 32

    def pop(self):
        assert len(self._state) > 0
        self._x, self._y, self._s, self._color, self._line_width = self._state.pop(-1)

    def line(self, points):
        dr = pygame.draw.lines(self.surface, self._color, False, self.tx_points(points), self._line_width)
        self._add_rect(dr)
        self._dirty = True
                
    def circle(self, center, radius):
        dr = pygame.draw.circle(self.surface, self._color, self.tx(center), int(round((radius*self._s))), self._line_width)
        self._add_rect(dr)
        self._dirty = True

    def rect(self, r):
        x,y = self.tx((r[0],r[1]))
        r = pygame.Rect(x, y, int(round((r[2]*self._s))), int(round((r[3]*self._s))))
        dr = pygame.draw.rect(self.surface, self._color, r, self._line_width)
        self._add_rect(dr)
        self._dirty = True

    def polygon(self, points):
        dr = pygame.draw.polygon(self.surface, self._color, self.tx_points(points), self._line_width)
        self._add_rect(dr)
        self._dirty = True

    def ellipse(self, r):
        x,y = self.tx((r[0],r[1]))
        r = pygame.Rect(x, y, int(round((r[2]*self._s))), int(round((r[3]*self._s))))
        dr = pygame.draw.ellipse(self.surface, self._color, r, self._line_width)
        self._add_rect(dr)
        self._dirty = True

    def blit(self, surface, xy):
        key = int(round(self._s * 10000))
        root = surface
        if key == 10000:
            pass
        else:
            surfaces = self._images.setdefault(surface,[])
            for s, root in surfaces:
                if s == key: 
                    break
            else:
                new_size = [int(round(i*self._s)) for i in surface.get_size()]
                root = pygame.transform.scale(surface, new_size)
                surfaces.append((key, root))
                if len(surfaces) > self.CACHE_SIZE: surfaces.pop(0)
        dr = self.surface.blit(root, self.tx(xy), None, self._blend_flags)
        self._add_rect(dr)
        self._dirty = True

    def curve(self, control, a=0.5, resolution=10):
        points = paths.hermite(control, a, resolution)
        points = self.tx_points(points)
        dr = pygame.draw.lines(self.surface, self._color, False, points, self._line_width)
        self._add_rect(dr)
        self._dirty = True

    


